#!/usr/bin/env python3
# Copyright 2019 The Bazel Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper tool to update changelog.md

Usage:
  update_changelog.py
  edit CHANGELOG.md
"""

import argparse
import io
import os
import re
import subprocess


def find_file_in_workspace(names, auto_create=True, max_depth=10):
  """Find a file within this WORKSPACE.

  Works up from the current directory to find a file of a given set of names.
  Stops at the WORKSPACE and creates it there if not found.

  Args:
    names: list of file names (for capitalizaiton);
  Walk up the tree from the current directory.  If there is none up through the
  finding WORKSPACE, create one next to workspace.

  Returns:
    path to changelog
  """
  dir_path = '.'
  for i in range(max_depth):
    for name in names:
      path = os.path.join(dir_path, name)
      if os.path.exists(path):
        return path
    if os.path.exists(os.path.join(dir_path, 'WORKSPACE')):
      if not auto_create:
        return None
      path = os.path.join(dir_path, names[0])
      with open(path, 'w') as x:
        x.write('\nInitial version\n')
      return path
    dir_path = os.path.join(dir_path, '..')
  # Your project is too broken to work. There is no good way to guess where to
  # create one.
  raise Exception('Could not find changelog.md within %s levels up.' %
                  max_depth)


def get_commit_log(from_tag):
  """Gets git commit log from a given tag."""

  cmd = ['git', 'log', '%s..HEAD' % from_tag]
  p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
  commit_log = p.stdout.read().decode('utf-8')
  p.wait()
  return commit_log


def guess_previous_release_tag():
  most_recent = None
  with subprocess.Popen(['git', 'tag'], stdout=subprocess.PIPE) as proc:
    # TODO(aiuto): This is crap. Do something better.
    most_recent = proc.stdout.read().decode('utf-8').split('\n')
    most_recent = [v for v in most_recent if v][-1]
  return most_recent


def lstrip_lines(text):
  return '\n'.join([line.strip() for line in text.strip().split('\n')])


def split_log(commit_log):
  """Splits a big commit log string into chunks by commit.
  
  Returns:
    dict: commit line => text
  """
  chunks = re.split(r'(^|\n)(commit [0-9a-f]+)\n', commit_log)
  ret = {}
  hash = None
  text = None
  for chunk in chunks:
    if chunk.startswith('commit '):
      if hash and text:
        ret[hash] = lstrip_lines(text)
      hash = chunk
      text = None
    else:
      text = (text or '') + chunk
  if hash and text:
    ret[hash] = lstrip_lines(text)
  return ret


def md_indent(text):
  """Does markdown indent on text."""
  lines = list(filter(lambda s: s, text.strip().split('\n')))
  return '-   ' + lines[0] + '\n    ' + '\n    '.join(lines[1:])


def merge_new_commits(change_log, new_commits, from_tag=None,
                      version='@TBD@'):
  with open(change_log, 'r') as f:
    change_log_content = f.read()
  existing_commits = split_log(change_log_content)
  ret = 'Release %s' % version
  ret += '\n\n**New Features**\n'
  for c, text in new_commits.items():
    if c not in existing_commits:
      print('Adding', c)
      ret += '-   ' + c + '\n\n' + md_indent(text) + '\n'
  ret += change_log_content
  return ret


def guess_current_version():
  version_file = find_file_in_workspace(names=['version.bzl'],
                                        auto_create=False)
  if not version_file:
    return None
  with open(version_file, 'r') as f:
    v_re = re.compile('^version *= *"([^"]*)"')
    for line in f:
      m = v_re.match(line)
      if m:
        return m.group(1)
  return None


def main():
  parser = argparse.ArgumentParser(description='update changelog.md')
  parser.add_argument(
        '--from_tag', default='auto',
        help='Gather diffs since this tag. Use "auto" to pick the tag of the'
             ' most recent release')
  options = parser.parse_args()

  from_tag = options.from_tag
  if from_tag == 'auto':
    from_tag = guess_previous_release_tag()
  print('From:', from_tag)
  new_commits = split_log(get_commit_log(from_tag))
  change_log = find_file_in_workspace(
      names=('CHANGELOG.MD', 'changlog.md', 'CHANGELOG.md'))
  version = guess_current_version()
  if version:
    print('Current Version:', version)
  else:
    print('No version.bzl file found. Can not guess current version.')

  new_log_content = merge_new_commits(change_log, new_commits, from_tag,
                                      version)

  with open(change_log, 'w') as f:
    f.write(new_log_content)
  print('Now edit %s to adjust the change log' % change_log)


if __name__ == '__main__':
  main()
